
/*
 * File:        TFT_keypad_BRL4.c
 * Author:      Bruce Land
 * Adapted from:
 *              main.c by
 * Author:      Syed Tahmid Mahbub
 * Target PIC:  PIC32MX250F128B
 */

// graphics libraries
#include "config.h"
#include "tft_master.h"
#include "tft_gfx.h"
// need for rand function
#include <stdlib.h>

// threading library
#include <plib.h>
// config.h sets 40 MHz
#define	SYS_FREQ 40000000
#include "pt_cornell_TFT.h"

//#include "Filter_bank_coeff_25.h"
#include "Filter_bandpass_14bit.h"

//===============================================================
typedef signed int fix16 ;
#define multfix16(a,b) ((fix16)(((( signed long long)(a))*(( signed long long)(b)))>>16)) //multiply two fixed 16:16
#define float2fix16(a) ((fix16)((a)*65536.0)) // 2^16
#define fix2float16(a) ((float)(a)/65536.0)
#define fix2int16(a)    ((int)((a)>>16))
#define int2fix16(a)    ((fix16)((a)<<16))
#define divfix16(a,b) ((fix16)((((signed long long)(a)<<16)/(b)))) 
#define sqrtfix16(a) (float2fix16(sqrt(fix2float16(a)))) 
#define absfix16(a) abs(a)

// lowpass coeff 
volatile fix16 alpha = float2fix16(0.005) ; // time constant ~ 200/8000 sec = 0.25 sec
volatile int current_key; 

// == bit fixed point 2.14 format ===============================
// resolution 2^-14 =  6.1035e-5
// dynamic range is +1.9999/-2.0
typedef signed short fix14 ;
#define multfix14(a,b) ((fix14)((((long)(a))*((long)(b)))>>14)) //multiply two fixed 2.14
#define float2fix14(a) ((fix14)((a)*16384.0)) // 2^14
#define fix2float14(a) ((float)(a)/16384.0)
#define absfix14(a) abs(a)
// Filter_bank_size  from Filter_bank_coeff.h
volatile fix14 history_1[Filter_bank_size], history_2[Filter_bank_size] ; //output history
volatile fix14 history_in1[Filter_bank_size], history_in2[Filter_bank_size];
volatile fix14 filter_out[Filter_bank_size];
//================================================================

/* Demo code for interfacing TFT (ILI9340 controller) to PIC32
 * The library has been modified from a similar Adafruit library
 */
// Adafruit data:
/***************************************************
  This is an example sketch for the Adafruit 2.2" SPI display.
  This library works with the Adafruit 2.2" TFT Breakout w/SD card
  ----> http://www.adafruit.com/products/1480

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

// string buffer
char buffer[60];

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_timer, pt_filter_bank, pt_key ;

// system 1 second interval tick
int sys_time_seconds ;

// === Timer Thread =================================================
// update a 1 second tick counter
static PT_THREAD (protothread_timer(struct pt *pt))
{
    PT_BEGIN(pt);
     tft_setCursor(0, 0);
     tft_setTextColor(ILI9340_WHITE);  tft_setTextSize(1);
     tft_writeString("Time in seconds since boot\n");
      while(1) {
        // yield time 1 second
        PT_YIELD_TIME_msec(1000) ;
        sys_time_seconds++ ;
        
        // draw sys_time
        //tft_fillRoundRect(0,10, 100, 14, 1, ILI9340_BLACK);// x,y,w,h,radius,color
        //tft_setCursor(0, 10);
        //tft_setTextColor(ILI9340_YELLOW); tft_setTextSize(2);
        //sprintf(buffer,"%d", ReadADC10(0));
        //tft_writeString(buffer);
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // timer thread

// === Color Thread =================================================
// draw 3 color patches for R,G,B from a random number
//static int color ;
//static int i;
static PT_THREAD (protothread_filter_bank(struct pt *pt))
{
    PT_BEGIN(pt);
      static short x=0, y=0, ypow, ypow_1[Filter_bank_size], color;
      static short color_index, display_phase ;
      // draw near y=0 or y=120
      display_phase = 0;
      
      while(1) {
        // yield time 1 second
        PT_YIELD_TIME_msec(12) ;
        
        //erase a little
        tft_fillRoundRect(x+1, 120*display_phase, 8, 100, 1, ILI9340_BLACK);// x,y,w,h,radius,color
        // draw spectrum 0-50 in y filter_out[i], ReadADC10(0)
        // stepping x every time
        for(y=0; y<Filter_bank_size; y++){
            //tft_drawCircle(short x0, short y0, short r, unsigned short color);
            //tft_drawCircle(y, ypos_1[y], 1, ILI9340_BLACK);
            ypow = filter_out[y]>>3; //0 to 1024
            
            /*
            if(ypow<10) color = ILI9340_BLUE ;
            else if(ypow<20) color = ILI9340_GREEN ;
            else if(ypow<50) color = ILI9340_YELLOW ;
            else  color = ILI9340_RED ;
              */
            
            // /*
            if (ypow<2) color_index = 0;
            else if (ypow<4) color_index = 0;
            else if (ypow<8) color_index = 2;
            else if (ypow<16) color_index = 4;
            else if (ypow<32) color_index = 8;
            else if (ypow<48) color_index = 12;
            else if (ypow<64) color_index = 16 ;
            else if (ypow<95) color_index = 20;
            else if (ypow<128) color_index = 24;
            else if (ypow<256) color_index = 26;
            else if (ypow<512) color_index = 31;
            else   color_index = 31;
            color = (color_index<<11) | (color_index<<6) | (color_index) ;
            if (color_index==31) color = ILI9340_RED;
            // */
            tft_fillRoundRect(x,100+120*display_phase-(y<<2), 1, 3, 1, color);// x,y,w,h,radius,color
            ypow_1[y] = ypow;
        }
        x++;
        if (x>319) {x=0 ;display_phase = !display_phase ;}
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // color thread

// === Keypad Thread =============================================
// connections:
// A0 -- row 1 -- thru 300 ohm resistor -- avoid short when two buttons pushed
// A1 -- row 2 -- thru 300 ohm resistor
// A2 -- row 3 -- thru 300 ohm resistor
// A3 -- row 4 -- thru 300 ohm resistor
// B7 -- col 1 -- 10k pulldown resistor -- avoid open circuit input when no button pushed
// B8 -- col 2 -- 10k pulldown resistor
// B9 -- col 3 -- 10k pulldown resistor

// PORT B
#define EnablePullDownB(bits) CNPUBCLR=bits; CNPDBSET=bits;
#define DisablePullDownB(bits) CNPDBCLR=bits;
#define EnablePullUpB(bits) CNPDBCLR=bits; CNPUBSET=bits;
#define DisablePullUpB(bits) CNPUBCLR=bits;
//PORT A
#define EnablePullDownA(bits) CNPUACLR=bits; CNPDASET=bits;
#define DisablePullDownA(bits) CNPDACLR=bits;
#define EnablePullUpA(bits) CNPDACLR=bits; CNPUASET=bits;
#define DisablePullUpA(bits) CNPUACLR=bits;

static PT_THREAD (protothread_key(struct pt *pt))
{
    PT_BEGIN(pt);
    static int keypad, i, pattern;
    // order is 0 thru 9 then * ==10 and # ==11
    // no press = -1
    // table is decoded to natural digit order (except for * and #)
    // 0x80 for col 1 ; 0x100 for col 2 ; 0x200 for col 3
    // 0x01 for row 1 ; 0x02 for row 2; etc
    static int keytable[12]={0x108, 0x81, 0x101, 0x201, 0x82, 0x102, 0x202, 0x84, 0x104, 0x204, 0x88, 0x208};
    // init the keypad pins A0-A3 and B7-B9
    // PortA ports as digital outputs
    mPORTASetPinsDigitalOut(BIT_0 | BIT_1 | BIT_2 | BIT_3);    //Set port as output
    // PortB as inputs
    mPORTBSetPinsDigitalIn(BIT_7 | BIT_8 | BIT_9);    //Set port as input
    // and turn on the pulldowns for these bits
    EnablePullDownB(BIT_7 | BIT_8 | BIT_9);
    //CNPDB = BIT_7 | BIT_8 | BIT_9; //0x0380 ;

      while(1) {

        // read each row sequentially
        mPORTAClearBits(BIT_0 | BIT_1 | BIT_2 | BIT_3);
        pattern = 1; mPORTASetBits(pattern);
        
        // yield time
        PT_YIELD_TIME_msec(30);
        //mPORTAClearBits(BIT_0 | BIT_1 | BIT_2 | BIT_3);
        //pattern = 1; mPORTASetBits(pattern);
        for (i=0; i<4; i++) {
            keypad  = mPORTBReadBits(BIT_7 | BIT_8 | BIT_9);
            if(keypad!=0) {keypad |= pattern ; break;}
            mPORTAClearBits(pattern);
            pattern <<= 1;
            mPORTASetBits(pattern);
        }

        // search for keycode
        if (keypad > 0){ // then button is pushed
            for (i=0; i<12; i++){
                if (keytable[i]==keypad) break;
            }
             current_key = i;
        }
        else i = -1; // no button pushed
       

        // draw key number
        //tft_fillRoundRect(30,200, 100, 28, 1, ILI9340_BLACK);// x,y,w,h,radius,color
        //tft_setCursor(30, 200);
        //tft_setTextColor(ILI9340_YELLOW); tft_setTextSize(4);
        //sprintf(buffer,"%d", current_key);
        //if (i==10)sprintf(buffer,"*");
        //if (i==11)sprintf(buffer,"#");
        //tft_writeString(buffer);

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // keypad thread

//== Timer 2 interrupt handler ===================================
// ipl2 means "interrupt priority level 2"
// rate is 8000/sec
void __ISR(_TIMER_2_VECTOR, ipl2) Timer2ISR(void)
{
    fix14 output;
    int i, input;
    // timing
    mPORTBSetPinsDigitalOut(BIT_13);
    mPORTBSetBits(BIT_13);
    // clear the interrupt flag
    mT2ClearIntFlag();
    // read the ADC
    input = (ReadADC10(0)-512)<<4;
    // read the first buffer position
    //channel0 = ReadADC10(0);   // read the result of channel 4 conversion from the idle buffer
    for(i=0; i<Filter_bank_size; i++){
     
        // second order butterworth computed by matlab butter(1,[lower upper])
        output = multfix14(b1[i], input-history_in2[i]) + multfix14(history_1[i], -a2[i]) +
                 multfix14(history_2[i], -a3[i]) ;
        //output = input; 
        history_2[i] = history_1[i];
        history_1[i] = output  ;
        history_in2[i] = history_in1[i];
        history_in1[i] = input ;
         
        // now lowpass the absolute value of the filter output
        // new_outout = alpha*input + (1-alpha)*last_output
        // time constant about 64 samples
        if (current_key==0)
            filter_out[i] = filter_out[i] + ((absfix14(output) - filter_out[i])>>6); 
        else if (current_key==1){
            if (output>filter_out[i]) filter_out[i] = output;
            else filter_out[i] = filter_out[i] + ((absfix14(output) - filter_out[i])>>6); 
        }
    }
    mPORTBClearBits(BIT_13);
}

/*//== Two-pole bandpass filter, 16-bit ============================
// coeff = {b1, -a2, -a3} noting that b1=-b3 b2=0 and a1=1
// history = { last_1_input, last_2_input, last_1_output, last_2_output}
fix16 IIR_butter_bp_2_16(fix16 input, fix16 *coeff2, fix16 *history2 )
{
    fix16 output;
    output = multfix16(input-history2[1], coeff2[0]) +
            multfix16(history2[2], coeff2[1]) +
            multfix16(history2[3], coeff2[2]) ;
            
    history2[1] = history2[0] ;
    history2[0] = input ;
    history2[3] = history2[2] ;
    history2[2] = output ;
    return output ;
}*/

// === Main  ======================================================
void main(void) {
 SYSTEMConfigPerformance(PBCLK);
  
  ANSELA = 0; ANSELB = 0; CM1CON = 0; CM2CON = 0;

  // the ADC ///////////////////////////////////////
        // configure and enable the ADC
	CloseADC10();	// ensure the ADC is off before setting the configuration

	// define  setup parameters for OpenADC10
	// Turn module on | ouput in integer | trigger mode auto | enable autosample
        // ADC_CLK_AUTO -- Internal counter ends sampling and starts conversion (Auto convert)
        // ADC_AUTO_SAMPLING_ON -- Sampling begins immediately after last conversion completes; SAMP bit is automatically set
        // ADC_AUTO_SAMPLING_OFF -- Sampling begins with AcquireADC10();
        #define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_ON//

	// define setup parameters for OpenADC10
	// ADC ref external  | disable offset test | disable scan mode | do 1 sample | use single buf | alternate mode off
	#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_OFF | ADC_SAMPLES_PER_INT_1 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF
        //
	// Define setup parameters for OpenADC10
        // use peripherial bus clock | set sample time | set ADC clock divider
        // ADC_CONV_CLK_Tcy2 means divide CLK_PB by 2 (max speed)
        // ADC_SAMPLE_TIME_5 seems to work with a source resistance < 1kohm
        // At PB clock 30 MHz, divide by two for ADC_CONV_CLK gives 66 nSec
        #define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_5 | ADC_CONV_CLK_Tcy2 //ADC_SAMPLE_TIME_15| ADC_CONV_CLK_Tcy2

	// define setup parameters for OpenADC10
	// set AN9 and  as analog inputs
	#define PARAM4	ENABLE_AN9_ANA

	// define setup parameters for OpenADC10
	// do not assign channels to scan
	#define PARAM5	SKIP_SCAN_ALL

	// use ground as neg ref for A | use AN9 for input A
	// configure to sample AN9 pin26
	SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN9 ); // configure to sample AN9
	OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using the parameters defined above

	EnableADC10(); // Enable the ADC

        // timer interrupt //////////////////////////
        // Set up timer2 on,  interrupts, internal clock, prescalar 1, toggle rate
        // at 40 MHz PB clock 
        // 5000 is 8 ksamples/sec
        OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, 5000);

        // set up the timer interrupt with a priority of 2
         ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
        mT2ClearIntFlag(); // and clear the interrupt flag
        
  // === config threads ==========
  // turns OFF UART support and debugger pin
  PT_setup();

  // === setup system wide interrupts  ========
  INTEnableSystemMultiVectoredInt();

  // init the threads
  PT_INIT(&pt_timer);
  PT_INIT(&pt_filter_bank);
  //PT_INIT(&pt_anim);
  //PT_INIT(&pt_key);

  // init the display
  tft_init_hw();
  tft_begin();
  tft_fillScreen(ILI9340_BLACK);
  //240x320 vertical display
  tft_setRotation(3); // Use tft_setRotation(1) for 320x240

  // seed random color
  srand(1);

  // round-robin scheduler for threads
  while (1){
      //PT_SCHEDULE(protothread_timer(&pt_timer));
      PT_SCHEDULE(protothread_filter_bank(&pt_filter_bank));
      //PT_SCHEDULE(protothread_anim(&pt_anim));
      PT_SCHEDULE(protothread_key(&pt_key));
      }
  } // main

// === end  ======================================================

